// Copyright (C) 2008
// by all contributors.

// This file is part of SCNSL.

// SCNSL is free software:
// you can redistribute it and/or modify it under the terms of the
// GNU Lesser General Public License as published by the Free Software
// Foundation, either version 3 of the License, or (at your option)
// any later version.

// SCNSL is distributed in the hope
// that it will be useful, but WITHOUT ANY WARRANTY;
// without even the implied warranty of MERCHANTABILITY or FITNESS
// FOR A PARTICULAR PURPOSE.
// See the GNU Lesser General Public License for more details.

// You should have received a copy of the GNU Lesser General Public
// License along with SCNSL,
// in a file named LICENSE.txt.
// If not, see <http://www.gnu.org/licenses/>.


#ifndef SCNSL__UTILS__COMMUNICATORSTACK_T_HH
#define SCNSL__UTILS__COMMUNICATORSTACK_T_HH



/// @file
/// Encompasses a stack of communicators.

#include "../core/Communicator_if_t.hh"

#include "CommunicatorBridge_t.hh"

namespace Scnsl { namespace Utils {

    /// @brief A stack of communicators.
    /// They are encompassed, to simplify instantation.
    ///
    /// This class does not holds the ownership of the encompassed communicators.
    /// If this is a wanted behaviour, register a communicator for deletion,
    /// calling the acquireOwnership() method.
    ///
    /// Design patterns:
    /// - Non copiable.
    /// - Non assignable.
    ///
    class CommunicatorStack_t:
        public Scnsl::Core::Communicator_if_t
    {
    public:

        /// @brief Constructor.
        ///
        /// @param head The highest leve communicator.
        /// @param tail The lowest level communicator.
        ///
        CommunicatorStack_t( Scnsl::Core::Communicator_if_t * head,
                             Scnsl::Core::Communicator_if_t * tail )
            throw ();

        /// @brief Virtual destructor.
        virtual
        ~CommunicatorStack_t()
        throw ();

        /// @brief Support methods.
        //@{

        /// @brief Acquires the ownership of the passed communicator.
        ///
        /// @param c The communicator.
        ///
        void acquireOwnership( Scnsl::Core::Communicator_if_t * c )
            throw ();

        //@}

        /// @name Interface methods.
        //@{

        virtual
        void setCarrier( const Scnsl::Core::Channel_if_t * ch, const carrier_t c )
            throw ( std::logic_error );

        virtual
        errorcode_t send( const Scnsl::Core::Packet_t & p )
            throw ( std::logic_error );

        virtual
        errorcode_t receive( const Scnsl::Core::Packet_t & p )
            throw ( std::logic_error );

/* TODO CHECK THIS
        virtual
        void bindTaskProxy( const Scnsl::Core::TaskProxy_if_t * tp,
                            Scnsl::Core::Communicator_if_t * c )
            throw ( std::logic_error );

        virtual
        void bindChannel( const Scnsl::Core::Channel_if_t * ch,
                          Scnsl::Core::Communicator_if_t * c )
            throw ( std::logic_error );

        virtual
        void stackUp( Scnsl::Core::Communicator_if_t * c )
            throw ( std::logic_error );

        virtual
        void stackDown( Scnsl::Core::Communicator_if_t * c )
            throw ( std::logic_error );
*/
        //@}


        /// @name Methods for interconnection with the internal bridge.
        //@{

        /// @brief The setCarrier() method called by the bridge.
        void bridgeSetCarrier( const Scnsl::Core::Channel_if_t * ch, const carrier_t c )
            throw ( std::logic_error );

        /// @brief The send() method called by the bridge.
        errorcode_t bridgeSend( const Scnsl::Core::Packet_t & p )
            throw ( std::logic_error );

        /// @brief The receive() method called by the bridge.
        errorcode_t bridgeReceive( const Scnsl::Core::Packet_t & p )
            throw ( std::logic_error );

        //@}

    protected:

        /// @brief The topmost communicator of the chain.
        Scnsl::Core::Communicator_if_t * _head;

        /// @brief The bottom communicator of the chain.
        Scnsl::Core::Communicator_if_t * _tail;

        /// @brief List of communicators owned.
        CommunicatorList_t _ownedCommunicators;

        /// @brief Internal communicator bridge.
        CommunicatorBridge_t _bridge;

    private:

        /// @brief Disabled copy constructor.
        CommunicatorStack_t( const CommunicatorStack_t & );

        /// @brief Disabled assignement operator.
        CommunicatorStack_t & operator = ( const CommunicatorStack_t & );
    };


  } }



#endif
