// Copyright (C) 2008
// by all contributors.

// This file is part of SCNSL.

// SCNSL is free software:
// you can redistribute it and/or modify it under the terms of the
// GNU Lesser General Public License as published by the Free Software
// Foundation, either version 3 of the License, or (at your option)
// any later version.

// SCNSL is distributed in the hope
// that it will be useful, but WITHOUT ANY WARRANTY;
// without even the implied warranty of MERCHANTABILITY or FITNESS
// FOR A PARTICULAR PURPOSE.
// See the GNU Lesser General Public License for more details.

// You should have received a copy of the GNU Lesser General Public
// License along with SCNSL,
// in a file named LICENSE.txt.
// If not, see <http://www.gnu.org/licenses/>.



/// @file
/// A TLM task.

#include <sstream>

#include "TlmTaskProxy_t.hh"
#include "TlmTask_if_t.hh"


using Scnsl::Tlm::TlmTask_if_t;


// ////////////////////////////////////////////////////////////////
// Destructor.
// ////////////////////////////////////////////////////////////////

TlmTask_if_t::~TlmTask_if_t()
    throw ()
{
    // Calling destructors for all sockets.
    // NOTE: reverse order not truly required here,
    // while it is a must in the STL.
    for ( register unsigned int i = _PROXIES_NUMBER; i != 0; )
    {
        --i;

        initiators[ i ].~InitiatorSocket_t();
        targets[ i ].~TargetSocket_t();
    }

    // Deallocating the memory.
    ::operator delete[] ( initiators );
    ::operator delete[] ( targets );
}


// ////////////////////////////////////////////////////////////////
// Constructor.
// ////////////////////////////////////////////////////////////////

TlmTask_if_t::TlmTask_if_t( sc_core::sc_module_name modulename,
                            const task_id_t id,
                            Scnsl::Core::Node_t * n,
                            const size_t proxies )
    throw ( std::invalid_argument ):
    // Parents:
    Scnsl::Core::Task_if_t( modulename, id, n ),
    tlm::tlm_fw_transport_if<>(),
    tlm::tlm_bw_transport_if<>(),
    // Sockets:
    initiators( NULL ),
    targets( NULL ),
    // Fields:
    _bounded_proxies( 0 ),
    _PROXIES_NUMBER( proxies ),
    _payload(),
    _delay( sc_core::SC_ZERO_TIME )
{
    if ( _PROXIES_NUMBER == 0 )
    {
        throw std::invalid_argument( "Invalid number of proxies." );
    }

    // Initializing in two steps, to be able to naming the sockets:

    initiators = static_cast< InitiatorSocket_t * >(
        operator new [] ( _PROXIES_NUMBER * sizeof( InitiatorSocket_t ) ));
    targets = static_cast< TargetSocket_t * >(
        operator new [] ( _PROXIES_NUMBER * sizeof( TargetSocket_t ) ));

    for ( register unsigned int i = 0; i < _PROXIES_NUMBER; ++i )
    {
        std::string s;
        std::stringstream ss;
        ss << i;
        ss >> s;

        // Calling sockets constructors.
        new (initiators + i) InitiatorSocket_t( ("initiator" + s).c_str() );
        new (targets + i) TargetSocket_t( ("target" + s).c_str() );

        // Binding:
        targets[ i ].bind( *this );
		initiators[ i ].bind( *this );
    }
}


bool TlmTask_if_t::get_direct_mem_ptr( tlm::tlm_generic_payload & p, tlm::tlm_dmi & dmi_data )
{
	return false;
    // To suppress compiler warnings:
    throw &p;
    throw &dmi_data;
}


tlm::tlm_sync_enum TlmTask_if_t::nb_transport_fw(tlm::tlm_generic_payload & p, tlm::tlm_phase& phase, sc_core::sc_time & t)
{
	return tlm::TLM_COMPLETED;
    // To suppress compiler warnings:
    throw &p;
    throw &phase;
    throw &t;
}


unsigned int TlmTask_if_t::transport_dbg(tlm::tlm_generic_payload & p)
{
	return 0;
    // To suppress compiler warnings:
    throw &p;
}


void TlmTask_if_t::invalidate_direct_mem_ptr( sc_dt::uint64 start_range, sc_dt::uint64 end_range )
{
    return;
    // To suppress compiler warnings:
    throw & start_range;
    throw & end_range;
}

tlm::tlm_sync_enum TlmTask_if_t::nb_transport_bw(tlm::tlm_generic_payload & p, tlm::tlm_phase& phase, sc_core::sc_time & t)
{
	return tlm::TLM_COMPLETED;
    // To suppress compiler warnings:
    throw &p;
    throw &phase;
    throw &t;
}


void TlmTask_if_t::bindTaskProxy( TlmTaskProxy_t * tp )
    throw ( std::logic_error )
{
    SCNSL_TRACE_DBG( 5, "<> bindTaskProxy().");

    if ( _bounded_proxies >= _PROXIES_NUMBER )
    {
        throw std::logic_error( "Already bounded all available proxies." );
    }

//     tp->initiator->bind( this->targets[ _bounded_proxies ] );
//     this->initiators[ _bounded_proxies ]->bind( tp->target );
//     tp->target->bind( this->initiators[ _bounded_proxies ] );
//     this->targets[ _bounded_proxies ]->bind( tp->initiator );
    tp->initiator( this->targets[ _bounded_proxies ] );
    this->initiators[ _bounded_proxies ]( tp->target );
    tp->setId( _bounded_proxies );

    ++ _bounded_proxies;
}



void TlmTask_if_t::send( const tlm_taskproxy_id_t tp, byte_t * buffer, const size_t s, const label_t label )
    throw ()
{
    SCNSL_TRACE_DBG( 5, "<> send().");
    SCNSL_TRACE_LOG( 3, "<> send().");

    _payload.set_command( PACKET_COMMAND );
    _payload.set_data_ptr( buffer );
    _payload.set_data_length( s );
    _payload.set_address( _ID );

	Scnsl::Core::data_extension_t * ext = new Scnsl::Core::data_extension_t;
	ext->label = label;
	_payload.set_extension( ext );

    initiators[ tp ]->b_transport( _payload, _delay );
}


// ////////////////////////////////////////////////////////////////
// Interface methods.
// ////////////////////////////////////////////////////////////////

TlmTask_if_t::size_t TlmTask_if_t::getProxiesNumber()
    const throw ()
{
    SCNSL_TRACE_DBG( 5, "<> getProxiesNumber().");
    return _PROXIES_NUMBER;
}

TlmTask_if_t::size_t TlmTask_if_t::getBoundedProxiesNumber()
    const throw ()
{
    SCNSL_TRACE_DBG( 5, "<> getBoundedProxiesNumber().");
    return _bounded_proxies;
}

